"""Defines Intel RAPL related classes."""
from typing import Callable

from .backend import Backend, backend
from .device import Device
from perun.units import Joule
from perun import log
import cpuinfo


@backend
class IntelRAPLBackend(Backend):
    """Intel RAPL as a source of cpu and memory devices.

    Uses pyRAPL to gather device information and creates metrics for each available device
    """

    name = "Intel RAPL"
    description = "Reads energy usage from CPUs and DRAM using Intel RAPL"

    def __init__(self) -> None:
        """Init IntelRAPLBackend."""
        super().__init__()
        log.debug("Initialized pyrapl")

    def setup(self):
        """Import pyrapl and gather basic information."""
        import pyRAPL

        self.sensor = pyRAPL.Sensor()
        cpuInfo = cpuinfo.get_cpu_info()
        self.cpu_name = cpuInfo["brand_raw"]
        self.thread_count = cpuInfo["count"]

    def close(self) -> None:
        """Backend shutdown code (does nothing for intel rapl)."""
        return

    def visibleDevices(self) -> set[str]:
        """
        Return string ids of visible devices.

        Returns:
            set[str]: Set with device string ids
        """
        visibleDevices = set()
        for device in self.sensor._available_devices:
            for socket in self.sensor._socket_ids:
                visibleDevices.add(f"{device.name}_{socket}")
        return visibleDevices

    def getDevices(self, deviceList: set[str]) -> list[Device]:
        """
        Gather devive objects based on a set of device ids.

        Args:
            deviceList (set[str]): Set containing devices ids

        Returns:
            list[Device]: Device objects
        """

        def getCallback(device_api, socket) -> Callable[[], float]:
            def func() -> float:
                return device_api.energy()[socket]

            return func

        for device, device_api in self.sensor._device_api.items():
            for socket in self.sensor._socket_ids:

                if f"{device.name}_{socket}" in deviceList:

                    deviceName = "CPU" if device.name == "PKG" else "DRAM"
                    name = f"{socket}_{deviceName}"
                    long_name = f"{self.cpu_name}:{socket}_{deviceName}"

                    self.devices.append(
                        Device(
                            name,
                            long_name,
                            Joule(),
                            "micro",
                            getCallback(device_api, socket),
                        )
                    )

        return self.devices


IntelRAPLBackend()



"""Device module."""
from dataclasses import dataclass
from typing import Callable, Any
from perun.units import Unit


@dataclass
class Device:
    """Defines a devices sensor properties."""

    id: str
    long_name: str
    unit: Unit
    mag: str
    measureCallback: Callable[[], float]

    def read(self) -> float:
        """Read value from sensor."""
        return self.measureCallback()

    def toDict(self) -> dict[str, Any]:
        """Return device as a dictionary."""
        return {
            "id": self.id,
            "long_name": self.long_name,
            "unit": self.unit,
            "mag": self.mag,
        }



"""Nvidia Mangement Library Source definition."""
from typing import Callable

from .backend import Backend, backend
from .device import Device
from perun.units import Watt
from perun import log
import pynvml
from pynvml import NVMLError


@backend
class NVMLSource(Backend):
    """NVMLSource class.

    Setups connection to NVML and creates relevant devices
    """

    name = "NVIDIA ML"
    description: str = "Access GPU information from NVML python bindings"

    def __init__(self) -> None:
        """Init NVIDIA ML Backend."""
        super().__init__()

    def close(self):
        """Backend shutdown code."""
        pynvml.nvmlShutdown()

    def visibleDevices(self) -> set[str]:
        """
        Return string ids of visible devices.

        Returns:
            set[str]: Set with device string ids
        """
        devices = set()
        for i in range(pynvml.nvmlDeviceGetCount()):
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            devices.add(pynvml.nvmlDeviceGetUUID(handle).decode("utf-8"))
        return devices

    def getDevices(self, deviceList: set[str]):
        """
        Gather device objects based on a set of device ids.

        Args:
            deviceList (set[str]): Set containing devices ids

        Returns:
            list[Device]: Device objects
        """

        def getCallback(handle) -> Callable[[], float]:
            def func() -> float:
                return pynvml.nvmlDeviceGetPowerUsage(handle)

            return func

        for device in deviceList:
            try:
                handle = pynvml.nvmlDeviceGetHandleByUUID(device.encode())

                name = f"CUDA:{device}"
                long_name = f"{pynvml.nvmlDeviceGetName(handle).decode('utf-8')}"

                self.devices.append(
                    Device(name, long_name, Watt(), "mili", getCallback(handle))
                )
            except NVMLError as e:
                log.debug(f"Could not find device {device}")
                log.debug(e)

        return self.devices

    def setup(self):
        """Init pynvml and gather number of devices."""
        pynvml.nvmlInit()
        deviceCount = pynvml.nvmlDeviceGetCount()
        log.debug(f"NVML Device count: {deviceCount}")


NVMLSource()



"""Storage Module."""
from pathlib import Path
from typing import Any, Union
from perun.backend.device import Device
from perun.units import MagnitudePrefix
from datetime import datetime
import h5py
from h5py import Group
from mpi4py.MPI import Comm
import numpy as np


class LocalStorage:
    """Object used to store device data inside the perunSubprocess."""

    def __init__(self, nodeName: str, devices: list[Device]):
        """Create dictionary based on nodeName and devices."""
        self.devices = [device.toDict() for device in devices]
        self.nodeName = nodeName
        self.nodeData: dict[str, list[Any]] = {"t_ns": []}
        for device in self.devices:
            self.nodeData[device["id"]] = []

    def addTimestep(self, timestamp: int, step: dict):
        """Add one step of information to storage."""
        self.nodeData["t_ns"].append(timestamp)
        for key, value in step.items():
            self.nodeData[key].append(value)

    def toDict(self) -> dict[str, Any]:
        """Lightweight local storage without the data."""
        return {
            "devices": self.devices,
            "nodeName": self.nodeName,
            "steps": len(self.nodeData["t_ns"]),
        }


class ExperimentStorage:
    """Store hardware measurments across multiple runs of the same experiment."""

    def __init__(self, filePath: Path, comm: Comm):
        """Initialize experiment storage."""
        self.comm = comm
        if self.comm.size > 1:
            self.file = h5py.File(filePath, "a", driver="mpio", comm=self.comm)
        else:
            self.file = h5py.File(filePath, "a")

        self.experimentName = self._getExperimentName(filePath)

        if self.experimentName not in self.file:
            rootGroup = self.file.create_group(self.experimentName)
            rootGroup.attrs["creation_date"] = str(datetime.utcnow())

    def _getExperimentName(self, filePath: Path) -> str:
        """Remove suffix from a file path."""
        return filePath.name.replace(filePath.suffix, "")

    def addExperimentRun(self, lStrg: Union[dict, None]) -> int:
        """Add new experiment group and setup the internal datasets."""
        expIdx: int = len(self.file[self.experimentName].keys())
        expGroup = self.file[self.experimentName].create_group(f"exp_{expIdx}")
        expGroup.attrs["experiment_date"] = str(datetime.utcnow())

        gatherdStrg: list[Union[dict, None]] = self.comm.allgather(lStrg)

        for strg in gatherdStrg:
            if strg:
                self._createNodeDataStrg(expGroup, strg)

        return expIdx

    def saveDeviceData(self, expId: int, lStrg: LocalStorage):
        """
        Write the device data on the newly created hdf5 datasets.

        Args:
            expId (int): id of the current experiment run
            lStrg (LocalStorage): Storage with hardware measurements
        """
        group = self.file[self.experimentName][f"exp_{expId}"][lStrg.nodeName]
        group["t_ns"][:] = np.array(lStrg.nodeData["t_ns"], dtype="float64")

        for device in lStrg.devices:
            dsId = self._dsFromDevice(device)
            group[dsId][:] = np.array(lStrg.nodeData[device["id"]])

    def _createNodeDataStrg(self, group: Group, strg: dict):
        """
        Create datasets for a mpi node.

        Args:
            group (Group): Parent hdf5 group
            strg (dict): Dictionary with node and device data
        """
        nodeGroup = group.create_group(strg["nodeName"])
        self._createTimestampDatabase(nodeGroup, strg)
        for device in strg["devices"]:
            self._createDataset(nodeGroup, device, strg["steps"])

    def _createTimestampDatabase(self, group: Group, strg: dict):
        """Initilize timestamp database."""
        t_ns_ds: h5py.Dataset = group.create_dataset(
            name="t_ns", shape=(strg["steps"],), dtype="float64"
        )
        t_ns_ds.attrs["long_name"] = "time"
        t_ns_ds.attrs["units"] = "seconds"
        t_ns_ds.attrs["mag"] = "nano"
        t_ns_ds.attrs["standard_name"] = "time"

    def _createDataset(self, group: Group, deviceDict: dict[str, Any], steps: int):
        """Create a dataset based on node name and device information."""
        ds: h5py.Dataset = group.create_dataset(
            self._dsFromDevice(deviceDict), shape=(steps,), dtype="float64"
        )
        ds.attrs["long_name"] = deviceDict["long_name"]
        ds.attrs["units"] = deviceDict["unit"].name
        ds.attrs["symbol"] = deviceDict["unit"].symbol
        ds.attrs.create("coordinates", data=["t_ns"])
        ds.attrs.create("valid_min", data=deviceDict["unit"].min_value, dtype="f")
        ds.attrs.create("valid_max", data=deviceDict["unit"].max_value, dtype="f")
        ds.attrs.create("_FillValue", data=-1.0, dtype="f")
        ds.attrs.create(
            "scale_factor", data=MagnitudePrefix.getFactor(deviceDict["mag"]), dtype="f"
        )
        ds.attrs["scale_prefix"] = deviceDict["mag"]
        ds.attrs.create("add_offset", data=1.0, dtype="f")

    def _dsFromDevice(self, device: dict[str, Any]) -> str:
        """
        Generate dataset name from device dict.

        Args:
            device (dict[str, any]): Device dict.

        Returns:
            str: Dataset name string.
        """
        prefixSym = MagnitudePrefix.getSymbol(device["mag"])
        return f"{device['id']}_{prefixSym}{device['unit'].symbol}"

    def close(self):
        """Close hdf5 file."""
        self.file.close()

    def getExperimentRun(self, index: int) -> Group:
        """
        Return experiment with the desired index.

        Args:
            index (int): Run index

        Returns:
            Group: h5py group with run info
        """
        if index == -1:
            index = len(self.file[self.experimentName].keys()) - 1
        return self.file[self.experimentName][f"exp_{index}"]

    def getExperimentRuns(self) -> list[Group]:
        """Return list of run hdf5 groups.

        Returns:
            list[Group]: List of hdf5 groups with run data
        """
        return self.file[self.experimentName].values()

    def getRootObject(self) -> Group:
        """
        Get hdf5 root object.

        Returns:
            Group: Root object
        """
        return self.file[self.experimentName]

    def toDict(self, data: bool = False) -> dict:
        """
        Create dict represation of storage object.

        Args:
            data (bool, optional): Include raw data. Defaults to False.

        Returns:
            dict: Storage dictionary
        """
        rootObj = self.getRootObject()
        expDict: dict[str, Any] = {"name": self.experimentName, "runs": []}
        for key, value in rootObj.attrs.items():
            expDict[key] = str(value)

        for run_id, runs in rootObj.items():
            expDict["runs"].append(self._runToDict(run_id, runs, data))

        return expDict

    def _runToDict(self, name: str, run: Group, data: bool = False) -> dict:
        """
        Get dictionary from run object.

        Args:
            name (str): Run name
            run (Group): Run object
            data (bool, optional): If raw data should be included. Defaults to False.

        Returns:
            dict: run dictionary
        """
        runDict: dict[str, Any] = {"id": name, "node": []}
        for key, value in run.attrs.items():
            runDict[key] = str(value)

        for node_id, node in run.items():
            runDict["node"].append(self._nodeToDict(node_id, node, data))
        return runDict

    def _nodeToDict(self, name: str, node: Group, data: bool = False) -> dict:
        """
        Transform h5py node to dictionary.

        Args:
            name (str): Node name
            node (Group): Node object
            data (bool, optional): If raw data should be included. Defaults to False.

        Returns:
            dict: Node dictionary
        """
        nodeDict: dict[str, Any] = {"id": name, "devices": []}

        for key, value in node.attrs.items():
            nodeDict[key] = str(value)

        for device_id, device in node.items():
            deviceDict: dict[str, Any] = {
                "id": device_id,
            }
            if data:
                deviceDict["data"] = device[:]

            for key, value in device.attrs.items():
                deviceDict[key] = str(value)

            nodeDict["devices"].append(deviceDict)

        return nodeDict



"""Core perun functionality."""
import time
from functools import reduce
import functools
from pathlib import Path
import platform
from typing import Optional
from perun.units import MagnitudePrefix
from perun.backend import Backend, Device
from perun.storage import ExperimentStorage, LocalStorage
from perun import log
from multiprocessing import Event, Process, Queue
from mpi4py import MPI
import perun
import sys
import h5py
import numpy as np


def getDeviceConfiguration(comm: MPI.Comm, backends: list[Backend]) -> list[str]:
    """
    Obtain a list with the assigned devices to the current rank.

    Args:
        comm (Comm): MPI communicator object
        backends (list[Backend]): List of available backend objects

    Returns:
        list[str]: IDs of assigned devices to the current rank
    """
    visibleDevices: set[str] = reduce(
        lambda a, b: a | b, [backend.visibleDevices() for backend in backends]
    )

    log.debug(f"Rank {comm.rank} : Visible devices {visibleDevices}")
    globalDeviceNames = comm.allgather(visibleDevices)
    globalHostnames = comm.allgather(platform.node())

    previousHosts = {}
    for index, (hostname, globalDevices) in enumerate(
        zip(globalHostnames, globalDeviceNames)
    ):
        if hostname not in previousHosts:
            previousHosts[hostname] = index
        else:
            prevIndex = previousHosts[hostname]
            globalDeviceNames[prevIndex] |= globalDevices
            globalDeviceNames[index] = set()

    return globalDeviceNames[comm.rank]


def monitor(frequency: float = 1.0, outDir: str = "./", format: str = "txt"):
    """Decorate function to monitor its energy usage."""

    def inner_function(func):
        @functools.wraps(func)
        def func_wrapper(*args, **kwargs):

            comm = MPI.COMM_WORLD
            start_event = Event()
            stop_event = Event()

            filePath: Path = Path(sys.argv[0])
            outPath: Path = Path(outDir)

            # Get node devices
            log.debug(f"Backends: {perun.backends}")
            lDeviceIds: list[str] = perun.getDeviceConfiguration(comm, perun.backends)

            for backend in perun.backends:
                backend.close()

            log.debug(f"Rank {comm.rank} - lDeviceIds : {lDeviceIds}")

            # If assigned devices, start subprocess
            if len(lDeviceIds) > 0:
                queue: Queue = Queue()
                perunSP = Process(
                    target=perun.perunSubprocess,
                    args=[queue, start_event, stop_event, lDeviceIds, frequency],
                )
                perunSP.start()
                start_event.wait()

            # Sync everyone
            comm.barrier()

            func(*args, **kwargs)
            stop_event.set()

            lStrg: Optional[LocalStorage]
            # Obtain perun subprocess results
            if len(lDeviceIds) > 0:
                perunSP.join()
                perunSP.close()
                lStrg = queue.get()
            else:
                lStrg = None

            # Sync
            comm.barrier()

            # Save raw data to hdf5

            if comm.rank == 0:
                if not outPath.exists():
                    outPath.mkdir(parents=True)

            scriptName = filePath.name.removesuffix(filePath.suffix)
            resultPath = outPath / f"{scriptName}.hdf5"
            log.debug(f"Result path: {resultPath}")
            expStrg = perun.ExperimentStorage(resultPath, comm)
            if lStrg:
                log.debug("Creating new experiment")
                expId = expStrg.addExperimentRun(lStrg.toDict())
                expStrg.saveDeviceData(expId, lStrg)
            else:
                expId = expStrg.addExperimentRun(None)

            # Post post-process
            perun.postprocessing(expStorage=expStrg)
            if comm.rank == 0:
                print(perun.report(expStrg, expIdx=expId, format=format))
            expStrg.close()

        return func_wrapper

    return inner_function


def perunSubprocess(
    queue: Queue, start_event, stop_event, deviceIds: set[str], frequency: float
):
    """
    Parallel function that samples energy values from hardware libraries.

    Args:
        queue (Queue): multiprocessing Queue object where the results are sent after finish
        start_event (_type_): Marks that perun finished setting up and started sampling from devices
        stop_event (_type_): Signal the subprocess that the monitored processed has finished
        deviceIds (set[str]): List of device ids to sample from
        frequency (int): Sampling frequency in Hz
    """
    from perun import backends

    lDevices: list[Device] = []
    for backend in backends:
        backend.setup()
        lDevices += backend.getDevices(deviceIds)

    log.debug(f"Perun SP lDevices: {lDevices}")
    lStrg = LocalStorage(platform.node(), lDevices)
    start_event.set()

    while not stop_event.wait(1.0 / frequency):
        t = time.time_ns()
        stepData = {}
        for device in lDevices:
            stepData[device.id] = device.read()
        lStrg.addTimestep(t, stepData)

    for backend in backends:
        backend.close()

    queue.put(lStrg)


def postprocessing(expStorage: ExperimentStorage):
    """Process the obtained data."""
    expRuns = expStorage.getExperimentRuns()
    totalExpEnergy_kJ = 0
    totalExpRuntime_s = 0
    for run in expRuns:
        if "totalRunEnergy_kJ" not in run.attrs:
            _postprocessRun(run)
        totalExpEnergy_kJ += run.attrs["totalRunEnergy_kJ"]
        totalExpRuntime_s += run.attrs["totalRunRuntime_s"]

    rootExp = expStorage.getRootObject()
    rootExp.attrs["totalExpEnergy_kJ"] = totalExpEnergy_kJ
    rootExp.attrs["totalExpRuntime_s"] = totalExpRuntime_s


def _postprocessRun(run: h5py.Group):
    """
    Collect run statistics and save them in the group.

    Args:
        run (h5py.Group): Run hdf5 group
    """
    totalRunEnergy_kJ = 0
    totalRunRuntime_s = 0
    runAvgPower_kW = 0
    totalRunRuntime_s = 0

    for node in run.values():
        if "totalNodeEnergy_kJ" not in node.attrs:
            _postprocessNode(node)
        totalRunEnergy_kJ += node.attrs["totalNodeEnergy_kJ"]
        runAvgPower_kW += node.attrs["avgNodePower_kW"]
        totalRunRuntime_s = max(totalRunRuntime_s, node.attrs["totalNodeRuntime_s"])

    run.attrs["totalRunEnergy_kJ"] = totalRunEnergy_kJ
    run.attrs["totalRunRuntime_s"] = totalRunRuntime_s
    run.attrs["runAvgPower_kW"] = runAvgPower_kW


def _postprocessNode(node: h5py.Group):
    """
    Collect node statistics and save them in the group.

    Args:
        node (h5py.Group): Node hdf5 group
    """
    totalNodeEnergy_kJ = 0
    avgNodePower_kW = 0
    t_ns_ds: h5py.Dataset = node["t_ns"]

    mag = t_ns_ds.attrs["mag"]
    transformFactor = MagnitudePrefix.transformFactor(mag, "")
    node.attrs["totalNodeRuntime_s"] = (t_ns_ds[-1] - t_ns_ds[0]) * transformFactor

    for device in node.values():
        if "/t_ns" not in device.name:
            if "totalDeviceEnergy_kJ" not in device.attrs:
                _postprocessDevice(device, t_ns_ds)
            totalNodeEnergy_kJ += device.attrs["totalDeviceEnergy_kJ"]
            avgNodePower_kW += device.attrs["avgDevicePower_kW"]

    node.attrs["totalNodeEnergy_kJ"] = totalNodeEnergy_kJ
    node.attrs["avgNodePower_kW"] = avgNodePower_kW


def _postprocessDevice(device: h5py.Dataset, t_ns: h5py.Dataset):
    """
    Process device raw data and gather report statistics.

    Args:
        device (h5py.Dataset): Device hdf5 dataset
        t_ns (h5py.Dataset): Timesteps hdf5 dataset
    """
    mag = device.attrs["scale_prefix"]
    mTransFactor = MagnitudePrefix.transformFactor(mag, "kilo")
    tTransFactor = MagnitudePrefix.transformFactor(t_ns.attrs["mag"], "")

    if device.attrs["units"] == "Joule":
        startEnergy = device[0]
        endEnergy = device[-1]
        if endEnergy < startEnergy:
            maxValue = np.finfo(device.dtype).max
            totalEnergy = maxValue - startEnergy + endEnergy
        else:
            totalEnergy = endEnergy - startEnergy
        totalDeviceEnergy_kJ = totalEnergy * mTransFactor
        device.attrs["totalDeviceEnergy_kJ"] = totalDeviceEnergy_kJ
        device.attrs["avgDevicePower_kW"] = totalDeviceEnergy_kJ / (
            (t_ns[-1] - t_ns[0]) * tTransFactor
        )
    elif device.attrs["units"] == "Watt":
        power_v = device[:] * mTransFactor
        t = t_ns[:] * tTransFactor
        energy_v = np.diff(t) * (power_v[:-1] + power_v[1:]) / 1
        energy_v = np.cumsum(energy_v)
        device.attrs["totalDeviceEnergy_kJ"] = energy_v[-1]
        device.attrs["avgDevicePower_kW"] = np.mean(power_v)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['perun', 'perun.backend']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'click>=8.1.3,<9.0.0',
 'h5py>=3.7.0,<4.0.0',
 'mpi4py>=3.1.3,<4.0.0',
 'py-cpuinfo>=8.0.0,<9.0.0',
 'pyRAPL>=0.2.3,<0.3.0',
 'pynvml>=11.4.1,<12.0.0',
 'python-dotenv>=0.20.0,<0.21.0']

entry_points = \
{'console_scripts': ['perun = perun.cli:cli']}

setup_kwargs = {
    'name': 'perun',
    'version': '0.1.0b6',
    'description': '',
    'long_description': '<div align="center">\n  <img src="https://raw.githubusercontent.com/Helmholtz-AI-Energy/perun/main/doc/images/perun.svg">\n</div>\n\nHave you ever wondered how much energy is used when training your neural network on the MNIST dataset? Want to get scared because of impact you are having on the evironment while doing "valuable" research? Are you interested in knowing how much carbon you are burning playing with DALL-E just to get attention on twitter? If the thing that was missing from your machine learning workflow was existential dread, this is the correct package for you!\n\n## Installation\n\nFrom PyPI:\n\n```$ pip install perun```\n\nFrom Github:\n\n```$ pip install git+https://github.com/Helmholtz-AI-Energy/perun```\n\n### Parallel h5py\n\nTo build h5py with mpi support:\n\n```CC=mpicc HDF5_MPI="ON" pip install --no-binary h5py h5py```\n\n## Usage\n\n### Command line\n\nTo get a quick report of the power usage of a python script simply run\n\n```$ perun monitor --format yaml path/to/your/script.py [args]```\n\nOr\n\n```$ python -m perun monitor --format json -o results/ path/to/your/script.py [args]```\n\n\n#### Subcommands\n\n**monitor**\n\nMonitor energy usage of a python script.\n\n```\nUsage: perun monitor [OPTIONS] SCRIPT [SCRIPT_ARGS]...\n\n  Gather power consumption from hardware devices while SCRIPT [SCRIPT_ARGS] is\n  running.\n\n  SCRIPT is a path to the python script to monitor, run with arguments\n  SCRIPT_ARGS.\n\nOptions:\n  -f, --frequency FLOAT         sampling frequency (in Hz)\n  --format [txt|yaml|yml|json]  report print format\n  -o, --outdir DIRECTORY        experiment data output directory\n  --help                        Show this message and exit.\n```\n\n**report**\n\nPrint a report from previous monitoring results.\n\n```\nUsage: perun report [OPTIONS] EXP_HDF5\n\n  Print consumption report from EXP_HDF5 on the command line on the desired\n  format.\n\n  EXP_HDF5 is an hdf5 file generated by perun after monitoring a script,\n  containing data gathered from hardware devices.\n\nOptions:\n  -f, --format [txt|yaml|yml|json]\n                                  report print format\n  --help                          Show this message and exit.\n```\n\n**postprocess**\n\nApply postprocessing to existing perun experiment data.\n\n```\nUsage: perun postprocess [OPTIONS] EXP_HDF5\n\n  Apply post-processing to EXP_HDF5 experiment file.\n\n  EXP_HDF5 is an hdf5 file generated by perun after monitoring a script,\n  containing data gathered from hardware devices.\n\nOptions:\n  --help  Show this message and exit.\n```\n\n### Decorator\n\nOr decorate the function that you want analysed\n\n```python\nimport perun\n\n@perun.monitor(outDir="results/", format="txt")\ndef training_loop(args, model, device, train_loader, test_loader, optimizer, scheduler):\n    for epoch in range(1, args.epochs + 1):\n        train(args, model, device, train_loader, optimizer, epoch)\n        test(model, device, test_loader)\n        scheduler.step()\n```\n\n\nOptional Arguments:\n\n|   |   |\n|---|---|\n|frequency: FLOAT             |sampling frequency (in Hz) |\n|format: [txt|yaml|yml|json]  |report print format |\n|outdir DIRECTORY:            |experiment data output directory |\n\n## Experiment data\n\nRaw data is saved in a hdf5 file, where results over multiple runs are accumulated.\n\nAt the top level, the root group containts groups for all individual runs, as well as information about creation date and total energy consumption over multiple runs.\n\nExperiments contain the information about a single run of the python script, and has information about nodes, devices per node, total runtime and avgerage power draw.\n\nThe same applies at the node level.\n\nAt the lowest group level, datasets with sample data from individual devices is collected, with the dataset attributes providing the device metadata, like id, measurement unit and magnitude, and possible value range.\n\n- exp_name (Root group)\n  - exp_0 (Group)\n    - node_name_0 (Group)\n      - device_0 (Dataset)\n        - units (Attribute)\n        - mag (Attribute)\n        - long_name (Attribute)\n        - ...\n      - device_1 (Dataset)\n      - ...\n    - node_name_1 (Group)\n    - ...\n  - exp_1 (Group)\n  - ...\n',
    'author': 'Gutiérrez Hermosillo Muriedas, Juan Pedro',
    'author_email': 'juanpedroghm@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Helmholtz-AI-Energy/perun',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""Backend module.

Defines backends and devices used for energy measurements. Connects to different backends and provides a unified interface for them.
"""
# flake8: noqa
from .backend import backends, Backend
from .device import Device
from . import intel_rapl
from . import nvml

__all__ = ["backends", "Device", "Backend"]
